@@ -1,3 +1,4 @@
+import logging
 import os
 import json
 import zlib
@@ -18,15 +19,17 @@ from .client_push import push_project_async, push_project_wait, push_project_fin
 from .utils import DateTimeEncoder
 from .version import __version__
 
-
 this_dir = os.path.dirname(os.path.realpath(__file__))
 
 
+class TokenError(Exception):
+    pass
+
+
 def decode_token_data(token):
     token_prefix = "Bearer ."
     if not token.startswith(token_prefix):
-        raise ValueError("Invalid token type")
-
+        raise TokenError(f"Token doesn't start with 'Bearer .': {token}")
     try:
         data = token[len(token_prefix):].split('.')[0]
         # add proper base64 padding"
@@ -34,28 +37,54 @@ def decode_token_data(token):
         decoded = zlib.decompress(base64.urlsafe_b64decode(data))
         return json.loads(decoded)
     except (IndexError, TypeError, ValueError, zlib.error):
-        raise ValueError("Invalid token data")
+        raise TokenError(f"Invalid token data: {token}")
 
 
 class MerginClient:
-    def __init__(self, url=None, auth_token=None, login=None, password=None, plugin_version=None):
-        self.url = url if url is not None else MerginClient.default_url()
+    """
+    Client for Mergin service.
 
+    :param url: String, Mergin service URL.
+    :param auth_token: String, Mergin authorization token.
+    :param login: String, login for Mergin service.
+    :param password: String, password for Mergin service.
+    :param plugin_version: String, info about plugin and QGIS version.
+    :param proxy_config: Dictionary, proxy settings to use when connecting to Mergin service. At least url and port
+        of the server should be provided. Expected keys: "url", "port", "user", "password".
+        Currently, only HTTP proxies are supported.
+    """
+    def __init__(self, url=None, auth_token=None, login=None, password=None, plugin_version=None, proxy_config=None):
+        self.url = url if url is not None else MerginClient.default_url()
         self._auth_params = None
         self._auth_session = None
         self._user_info = None
         self.client_version = "Python-client/" + __version__
         if plugin_version is not None:   # this could be e.g. "Plugin/2020.1 QGIS/3.14"
             self.client_version += " " + plugin_version
+        self.setup_logging()
         if auth_token:
-            token_data = decode_token_data(auth_token)
-            self._auth_session = {
-                "token": auth_token,
-                "expire": dateutil.parser.parse(token_data["expire"])
-            }
-            self._user_info = {
-                "username": token_data["username"]
-            }
+            try:
+                token_data = decode_token_data(auth_token)
+                self._auth_session = {"token": auth_token, "expire": dateutil.parser.parse(token_data["expire"])}
+                self._user_info = {"username": token_data["username"]}
+            except TokenError as e:
+                self.log.error(e)
+                raise ClientError("Auth token error: " + str(e))
+        handlers = []
+
+        # Create handlers for proxy, if needed
+        if proxy_config is not None:
+            proxy_url_parsed = urllib.parse.urlparse(proxy_config["url"])
+            proxy_url = proxy_url_parsed.path
+            if proxy_config["user"] and proxy_config["password"]:
+                handlers.append(
+                    urllib.request.ProxyHandler(
+                        {"https": f"{proxy_config['user']}:{proxy_config['password']}@{proxy_url}:{proxy_config['port']}"}
+                    )
+                )
+                handlers.append(urllib.request.HTTPBasicAuthHandler())
+            else:
+                handlers.append(urllib.request.ProxyHandler({"https": f"{proxy_url}:{proxy_config['port']}"}))
 
         # fix for wrong macos installation of python certificates,
         # see https://github.com/lutraconsulting/qgis-mergin-plugin/issues/70
@@ -63,7 +92,7 @@ class MerginClient:
         # is fixed.
         default_capath = ssl.get_default_verify_paths().openssl_capath
         if os.path.exists(default_capath):
-            self.opener = urllib.request.build_opener()
+            self.opener = urllib.request.build_opener(*handlers, urllib.request.HTTPSHandler())
         else:
             cafile = os.path.join(this_dir, 'cert.pem')
             if not os.path.exists(cafile):
@@ -71,7 +100,7 @@ class MerginClient:
             ctx = ssl.SSLContext()
             ctx.load_verify_locations(cafile)
             https_handler = urllib.request.HTTPSHandler(context=ctx)
-            self.opener = urllib.request.build_opener(https_handler)
+            self.opener = urllib.request.build_opener(*handlers, https_handler)
         urllib.request.install_opener(self.opener)
 
         if login and not password:
@@ -87,10 +116,25 @@ class MerginClient:
             if not self._auth_session:
                 self.login(login, password)
 
+    def setup_logging(self):
+        """Setup Mergin client logging."""
+        client_log_file = os.environ.get('MERGIN_CLIENT_LOG', None)
+        self.log = logging.getLogger('mergin.client')
+        self.log.setLevel(logging.DEBUG)  # log everything (it would otherwise log just warnings+errors)
+        if not self.log.handlers:
+            if client_log_file:
+                log_handler = logging.FileHandler(client_log_file)
+                log_handler.setFormatter(logging.Formatter('%(asctime)s %(message)s'))
+                self.log.addHandler(log_handler)
+            else:
+                # no Mergin log path in the environment - create a null handler that does nothing
+                null_logger = logging.NullHandler()
+                self.log.addHandler(null_logger)
+
     @staticmethod
     def default_url():
         """ Returns URL of the public instance of Mergin """
-        return 'https://public.cloudmergin.com'
+        return "https://public.cloudmergin.com"
 
     def user_agent_info(self):
         """ Returns string as it is sent as User-Agent http header to the server """
@@ -108,24 +152,27 @@ class MerginClient:
         return f"{self.client_version} ({platform.system()}/{system_version})"
 
     def _check_token(f):
+        """Wrapper for creating/renewing authorization token."""
         def wrapper(self, *args):
-            if (not self._auth_session or self._auth_session['expire'] < datetime.now(timezone.utc)) and self._auth_params:
-                self.login(self._auth_params['login'], self._auth_params['password'])
+            if self._auth_params:
+                if self._auth_session:
+                    # Refresh auth token if it expired or will expire very soon
+                    delta = self._auth_session["expire"] - datetime.now(timezone.utc)
+                    if delta.total_seconds() < 5:
+                        self.log.info("Token has expired - refreshing...")
+                        self.login(self._auth_params["login"], self._auth_params["password"])
+                else:
+                    # Create a new authorization token
+                    self.log.info(f"No token - login user: {self._auth_params['login']}")
+                    self.login(self._auth_params['login'], self._auth_params['password'])
             return f(self, *args)
         return wrapper
 
     @_check_token
     def _do_request(self, request):
+        """General server request method."""
         if self._auth_session:
-            delta = self._auth_session["expire"] - datetime.now(timezone.utc)
-            if delta.total_seconds() < 1:
-                self._auth_session = None
-                # Refresh auth token when login credentials are available
-                if self._auth_params:
-                    self.login(self._auth_params["login"], self._auth_params["password"])
-
-            if self._auth_session:
-                request.add_header("Authorization", self._auth_session["token"])
+            request.add_header("Authorization", self._auth_session["token"])
         request.add_header("User-Agent", self.user_agent_info())
         try:
             return self.opener.open(request)
@@ -214,6 +261,8 @@ class MerginClient:
             "login": login,
             "password": password
         }
+        self._auth_session = None
+        self.log.info(f"Going to log in user {login}")
         try:
             self._auth_params = params
             url = urllib.parse.urljoin(self.url, urllib.parse.quote("/v1/auth/login"))
@@ -226,11 +275,13 @@ class MerginClient:
         except urllib.error.HTTPError as e:
             if e.headers.get("Content-Type", "") == "application/problem+json":
                 info = json.load(e)
+                self.log.info(f"Login problem: {info.get('detail')}")
                 raise LoginError(info.get("detail"))
+            self.log.info(f"Login problem: {e.read().decode('utf-8')}")
             raise LoginError(e.read().decode("utf-8"))
         except urllib.error.URLError as e:
             # e.g. when DNS resolution fails (no internet connection?)
-            raise ClientError("Error trying to log in: " + str(e))
+            raise ClientError("failure reason: " + str(e.reason))
         self._auth_session = {
             "token": "Bearer %s" % session["token"],
             "expire": dateutil.parser.parse(session["expire"])
@@ -238,6 +289,7 @@ class MerginClient:
         self._user_info = {
             "username": data["username"]
         }
+        self.log.info(f"User {data['username']} successfully logged in.")
         return session
 
     def username(self):
@@ -454,7 +506,6 @@ class MerginClient:
             detail = f"Project path: {project_path}"
             raise ClientError(str(e), detail)
 
-
     def push_project(self, directory):
         """
         Upload local changes to the repository.
