@@ -1,3 +1,4 @@
+import logging
 import os
 import tempfile
 import shutil
@@ -5,7 +6,8 @@ from datetime import datetime, timedelta
 import pytest
 import pytz
 
-from ..client import MerginClient, ClientError, MerginProject, LoginError
+from ..client import MerginClient, ClientError, MerginProject, LoginError, decode_token_data, TokenError
+from ..client_push import push_project_async, push_project_cancel
 from ..utils import generate_checksum
 
 SERVER_URL = os.environ.get('TEST_MERGIN_URL')
@@ -57,8 +59,13 @@ def test_login(mc):
     token = mc._auth_session['token']
     assert MerginClient(mc.url, auth_token=token)
 
-    with pytest.raises(ValueError, match='Invalid token data'):
-        MerginClient(mc.url, auth_token='Bearer .jas646kgfa')
+    invalid_token = "Completely invalid token...."
+    with pytest.raises(TokenError, match=f"Token doesn't start with 'Bearer .': {invalid_token}"):
+        decode_token_data(invalid_token)
+
+    invalid_token = "Bearer .jas646kgfa"
+    with pytest.raises(TokenError, match=f"Invalid token data: {invalid_token}"):
+        decode_token_data(invalid_token)
 
     with pytest.raises(LoginError, match='Invalid username or password'):
         mc.login('foo', 'bar')
@@ -211,6 +218,51 @@ def test_push_pull_changes(mc):
     assert generate_checksum(os.path.join(project_dir_2, f_updated)) == f_remote_checksum
 
 
+def test_cancel_push(mc):
+    """
+    Start pushing and cancel the process, then try to push again and check if previous sync process was cleanly
+    finished.
+    """
+    test_project = "test_cancel_push"
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project + "_3")  # primary project dir for updates
+    project_dir_2 = os.path.join(TMP_DIR, test_project + "_4")
+    cleanup(mc, project, [project_dir, project_dir_2])
+    # create remote project
+    shutil.copytree(TEST_DATA_DIR, project_dir)
+    mc.create_project_and_push(test_project, project_dir)
+
+    # modify the project (add, update)
+    f_added = 'new.txt'
+    with open(os.path.join(project_dir, f_added), 'w') as f:
+        f.write("new file")
+    f_updated = "test3.txt"
+    modification = "Modified"
+    with open(os.path.join(project_dir, f_updated), 'w') as f:
+        f.write(modification)
+
+    # check changes before applied
+    pull_changes, push_changes, _ = mc.project_status(project_dir)
+    assert not sum(len(v) for v in pull_changes.values())
+    assert next((f for f in push_changes['added'] if f['path'] == f_added), None)
+    assert next((f for f in push_changes['updated'] if f['path'] == f_updated), None)
+
+    # start pushing and then cancel the job
+    job = push_project_async(mc, project_dir)
+    push_project_cancel(job)
+
+    # if cancelled properly, we should be now able to do the push without any problem
+    mc.push_project(project_dir)
+
+    # download the project to a different directory and check the version and content
+    mc.download_project(project, project_dir_2)
+    mp = MerginProject(project_dir_2)
+    assert mp.metadata["version"] == 'v2'
+    assert os.path.exists(os.path.join(project_dir_2, f_added))
+    with open(os.path.join(project_dir_2, f_updated), 'r') as f:
+        assert f.read() == modification
+
+
 def test_ignore_files(mc):
     test_project = 'test_blacklist'
     project = API_USER + '/' + test_project
@@ -355,10 +407,21 @@ def test_list_of_push_changes(mc):
     pull_changes, push_changes, push_changes_summary = mc.project_status(project_dir)
     assert str(push_changes_summary) == PUSH_CHANGES_SUMMARY
 
+
+def test_token_renewal(mc):
+    """Test token regeneration in case it has expired."""
+    test_project = 'test_token_renewal'
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project)  # primary project dir for updates
+
+    cleanup(mc, project, [project_dir])
+    shutil.copytree(TEST_DATA_DIR, project_dir)
+    mc.create_project_and_push(test_project, project_dir)
+
     mc._auth_session["expire"] = datetime.now().replace(tzinfo=pytz.utc) - timedelta(days=1)
-    mc._auth_params = None
-    with pytest.raises(ClientError, match="You do not have permissions for this project"):
-        mc.project_status(project_dir)
+    pull_changes, push_changes, _ = mc.project_status(project_dir)
+    to_expire = mc._auth_session["expire"] - datetime.now().replace(tzinfo=pytz.utc)
+    assert to_expire.total_seconds() > (9 * 3600)
 
 
 def test_force_gpkg_update(mc):
@@ -736,3 +799,53 @@ def test_paginated_project_list(mc):
     assert len(projects) == 2
     for i, project in enumerate(projects):
         assert project["name"] == sorted_test_names[i+2]
+
+
+def test_missing_local_file_pull(mc):
+    """Test pull of a project where a file deleted in the service is missing for some reason."""
+
+    test_project = "test_dir"
+    file_to_remove = "test2.txt"
+    project = API_USER + '/' + test_project
+    project_dir = os.path.join(TMP_DIR, test_project + "_5")  # primary project dir for updates
+    project_dir_2 = os.path.join(TMP_DIR, test_project + "_6")  # concurrent project dir
+    test_data_dir = os.path.join(os.path.dirname(os.path.realpath(__file__)), "test_data", test_project)
+
+    cleanup(mc, project, [project_dir, project_dir_2])
+    # create remote project
+    shutil.copytree(test_data_dir, project_dir)
+    mc.create_project_and_push(test_project, project_dir)
+
+    # remove a file in a different directory
+    mc.download_project(project, project_dir_2)
+    os.remove(os.path.join(project_dir_2, file_to_remove))
+    mc.push_project(project_dir_2)
+
+    # manually remove the file also locally in the original project dir
+    os.remove(os.path.join(project_dir, file_to_remove))
+
+    # try to sync again  -- it should not crash
+    mc.pull_project(project_dir)
+    mc.push_project(project_dir)
+
+
+def test_logging(mc):
+    """Test logger creation with and w/out the env variable for log file."""
+    assert isinstance(mc.log.handlers[0], logging.NullHandler)
+    mc.log.info("Test info log...")
+    # remove the Null handler and set the env variable with the global log file path
+    mc.log.handlers = []
+    os.environ['MERGIN_CLIENT_LOG'] = os.path.join(TMP_DIR, 'global-mergin-log.txt')
+    assert os.environ.get('MERGIN_CLIENT_LOG', None) is not None
+    token = mc._auth_session['token']
+    mc1 = MerginClient(mc.url, auth_token=token)
+    assert isinstance(mc1.log.handlers[0], logging.FileHandler)
+    mc1.log.info("Test log info to the log file...")
+    # cleanup
+    mc.log.handlers = []
+    del os.environ['MERGIN_CLIENT_LOG']
+
+
+def test_server_compatibility(mc):
+    """Test server compatibility."""
+    assert mc.is_server_compatible()
